

#include <windows.h>
#include "cv.h"
#include "highgui.h"
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include "filter.h"

#include <io.h>
#include <string>
#include <vector>
#include <map>
#include <set>

#ifndef _DEBUG
#pragma comment(lib,"cv.lib")
#pragma comment(lib,"cvaux.lib")
#pragma comment(lib,"cxcore.lib")
#pragma comment(lib,"highgui.lib")
#else
#pragma comment(lib,"cvd.lib")
#pragma comment(lib,"cvauxd.lib")
#pragma comment(lib,"cxcored.lib")
#pragma comment(lib,"highguid.lib")
#endif

#define YC2R(Y,Cb,Cr) (Y + 1.402*Cr)
#define YC2G(Y,Cb,Cr) (Y - 0.344*Cb - 0.714*Cr)
#define YC2B(Y,Cb,Cr) (Y + 1.772*Cb)
#define RGB2Y(R,G,B) (0.299*R + 0.587*G + 0.114*B)
#define RGB2CB(R,G,B) (-0.169*R - 0.331*G + 0.500*B)
#define RGB2CR(R,G,B) (0.500*R - 0.419*G - 0.081*B)

// [min, max] ͈̔͂ɖOa
template <class T>
T Saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

template <>
int Saturate(int v, int min, int max) {
    v -= min;
    v &= ~(v >> 31);
    v += min;

    v -= max;
    v &=  (v >> 31);
    v += max;
    
    return v;
}


void CopyBufToCvBGRImg(
		PIXEL_YC *buf, IplImage *img,
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int src,ix0,iy0,ix1,iy1;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		src = x0 + iy0 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			int y  = buf[src].y;
			int cb = buf[src].cb;
			int cr = buf[src].cr;

			int r = ( 255*y + ( ((22881*cr>>16)+3) <<10) ) >>12;
			int g = ( 255*y + ( ((-5616*cb>>16)+(-11655*cr>>16)+3) <<10) ) >>12;
			int b = ( 255*y + ( ((28919*cb>>16)+3) <<10) ) >>12;

			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+0) = Saturate(b, 0, 255);
			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+1) = Saturate(g, 0, 255);
			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+2) = Saturate(r, 0, 255);
			src++;
			ix1++;
		}
	}
}

void CopyBufToCvGrayImg(
		PIXEL_YC *buf, IplImage *img,
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int src,ix0,iy0,ix1,iy1;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		src = x0 + iy0 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			int y  = buf[src].y / 16;
			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1) = (unsigned char)Saturate(y, 0, 255);
			src++;
			ix1++;
		}
	}
}


void CopyCvBGRImgToBuf(
		IplImage *img, PIXEL_YC *buf, 
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int dst,ix0,iy0,ix1,iy1;
	PIXEL_YC yc;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		dst = x1 + iy1 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			int r,g,b;
			b = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+0);
			g = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+1);
			r = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+2);
			yc.y =  ((4918*r+354)>>10) + ((9655*g+585)>>10) + ((1875*b+523)>>10);
			yc.cb = ((-2775*r+240)>>10) + ((-5449*g+515)>>10) + (( 8224*b+256)>>10);
			yc.cr = ((8224*r+256)>>10) + ((-6887*g+110)>>10) + ((-1337*b+646)>>10);

			buf[dst] = yc;
			dst++;
			ix1++;
		}
	}
}


static inline int is_neg(int v) {
    return (((unsigned int)v) >> (sizeof(v)*8 - 1));
}

static inline int modulo(int a, int b) {
	int m = a % b;
//	return m + b * is_neg(m);

	if (m >= 0) {
		return m;
	} else {
		return m + b;
	}
}

typedef union {
	int64 i;
	struct {
		float f1;
		float f2;
	} f;
} INT64FLOAT2;

static inline int64 int64f2(float f1, float f2) {
	INT64FLOAT2 v;
	v.f.f1 = f1;
	v.f.f2 = f2;
	return v.i;
}


static IplImage *img0 = NULL;
static IplImage *img1 = NULL;
static IplImage *img_mask = NULL;
static IplImage *img_smask = NULL;
static IplImage *img_bg = NULL;

static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;
static std::vector<int> g_bef_check;

static CvSubdiv2D *g_subdiv;
static CvMemStorage *g_subdivStorage;

static int g_nDiv = 1000;
static int g_lineType = CV_AA;
static int g_lineL = 0;
static CvScalar g_lineColor = CV_RGB(0,0,0); 
static CvRNG g_rngState;

static double g_equality = 1.0;
static bool g_divEqual = false;


static int getTrackVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_trackMap.find(str);

	if (itr == g_trackMap.end()) {
		sprintf_s(errmsg, "track %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->track[itr->second];
}

static int getCheckVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_checkMap.find(str);

	if (itr == g_checkMap.end()) {
		sprintf_s(errmsg, "check %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->check[itr->second];
}


static void createImages(CvSize size) {
	img0		= cvCreateImage(size, IPL_DEPTH_8U, 3);
	img1		= cvCreateImage(size, IPL_DEPTH_8U, 3);
	img_mask	= cvCreateImage(size, IPL_DEPTH_8U, 1);
	img_smask	= cvCreateImage(size, IPL_DEPTH_8U, 1);
	img_bg		= cvCreateImage(size, IPL_DEPTH_8U, 3);
}

static void releaseImages() {
	cvReleaseImage(&img0);
	cvReleaseImage(&img1);
	cvReleaseImage(&img_mask);
	cvReleaseImage(&img_smask);
	cvReleaseImage(&img_bg);
}


void initVoronoi(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}
	g_bef_check.resize(fdp->check_n, 0);
}

void updateVoronoi(FILTER *fp) {
}


void exitVoronoi() {
	releaseImages();
}


static CvSubdiv2D* init_delaunay( CvMemStorage* storage,
                           CvRect rect )
{
    CvSubdiv2D* subdiv;

    subdiv = cvCreateSubdiv2D( CV_SEQ_KIND_SUBDIV2D, sizeof(*subdiv),
                               sizeof(CvSubdiv2DPoint),
                               sizeof(CvQuadEdge2D),
                               storage );
    cvInitSubdivDelaunay2D( subdiv, rect );

    return subdiv;
}


static void draw_subdiv_point( IplImage* img, CvPoint2D32f fp, CvScalar color )
{
    cvCircle( img, cvPoint(cvRound(fp.x), cvRound(fp.y)), 3, color, CV_FILLED, 8, 0 );
}


static void draw_subdiv_edge( IplImage* img, CvSubdiv2DEdge edge, CvScalar color )
{
    CvSubdiv2DPoint* org_pt;
    CvSubdiv2DPoint* dst_pt;
    CvPoint2D32f org;
    CvPoint2D32f dst;
    CvPoint iorg, idst;

    org_pt = cvSubdiv2DEdgeOrg(edge);
    dst_pt = cvSubdiv2DEdgeDst(edge);

    if( org_pt && dst_pt )
    {
        org = org_pt->pt;
        dst = dst_pt->pt;

        iorg = cvPoint( cvRound( org.x ), cvRound( org.y ));
        idst = cvPoint( cvRound( dst.x ), cvRound( dst.y ));

        cvLine( img, iorg, idst, color, 1, CV_AA, 0 );
    }
}


static void draw_subdiv( IplImage* img, CvSubdiv2D* subdiv,
                  CvScalar delaunay_color, CvScalar voronoi_color )
{
    CvSeqReader  reader;
    int i, total = subdiv->edges->total;
    int elem_size = subdiv->edges->elem_size;

    cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );

    for( i = 0; i < total; i++ )
    {
        CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);

        if( CV_IS_SET_ELEM( edge ))
        {
            draw_subdiv_edge( img, (CvSubdiv2DEdge)edge + 1, voronoi_color );
            draw_subdiv_edge( img, (CvSubdiv2DEdge)edge, delaunay_color );
        }

        CV_NEXT_SEQ_ELEM( elem_size, reader );
    }
}

static void draw_subdiv_facet(IplImage *src, IplImage *dst, CvSubdiv2DEdge edge, bool doFill, bool doDrawLine)
{
    CvSubdiv2DEdge t = edge;
    int i, count = 0;
    //CvPoint* buf = 0;
	CvPoint ptbuf[255];
	CvPoint *buf = ptbuf;
	CvPoint2D32f cp = {0.0, 0.0};

    // count number of edges in facet
    do
    {
        count++;
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    } while (t != edge );

    //buf = (CvPoint*)malloc( count * sizeof(buf[0]));
	if (count > 255) {
		return;
	}

    // gather points
    t = edge;
    for( i = 0; i < count; i++ )
    {
        CvSubdiv2DPoint* pt = cvSubdiv2DEdgeOrg( t );
        if( !pt ) break;

		cp.x += pt->pt.x;
		cp.y += pt->pt.y;

		buf[i] = cvPoint( cvRound(pt->pt.x), cvRound(pt->pt.y));
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    }

    if( i == count ) {
		cp.x /= count;
		cp.y /= count;

		int y = Saturate(cvRound(cp.y), 0, src->height - 1);
		int x = Saturate(cvRound(cp.x), 0, src->width  - 1);

		CvScalar col = CV_RGB(CV_IMAGE_ELEM(src, BYTE, y, x*3+2), 
							  CV_IMAGE_ELEM(src, BYTE, y, x*3+1),
							  CV_IMAGE_ELEM(src, BYTE, y, x*3+0));

		if (doFill) {
			cvFillConvexPoly( dst, buf, count, col, g_lineType, 0 );
		}
		if (doDrawLine) {
			cvPolyLine( dst, &buf, &count, 1, 1, g_lineColor, 1, g_lineType, 0);
		}
		/*
		if (e == 3) {
			CvPoint2D32f p = pt->pt;
			p.x += 5;
			draw_subdiv_point( dst, p, CV_RGB(0,255,255));
			p.x -= 5;
		} else {
			draw_subdiv_point( dst, pt->pt, CV_RGB(0,0,0));
		}
		*/
		//draw_subdiv_point( dst, cp, CV_RGB(0,255,255));
    }
    //free( buf );
}

static void drawSubdivVoronoi(IplImage *src, IplImage *dst, CvSubdiv2D* subdiv, bool doFill = true, bool doDrawLine = true)
{
    CvSeqReader  reader;
    int i, total = subdiv->edges->total;
    int elem_size = subdiv->edges->elem_size;
	std::set<int64> drawnSet;

    cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );

    for( i = 0; i < total; i++ )
    {
        CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);

        if( CV_IS_SET_ELEM( edge ))
        {
            CvSubdiv2DEdge e = (CvSubdiv2DEdge)edge;
			
			CvPoint2D32f sp = cvSubdiv2DEdgeOrg(e)->pt;
			CvPoint2D32f ep = cvSubdiv2DEdgeDst(e)->pt;

	        // left
			if (drawnSet.find(int64f2(sp.x, sp.y)) == drawnSet.end()) {
				drawnSet.insert(int64f2(sp.x, sp.y));
	            draw_subdiv_facet( src, dst, cvSubdiv2DRotateEdge( e, 1 ), doFill, doDrawLine);
			}

            // right
			if (drawnSet.find(int64f2(ep.x, ep.y)) == drawnSet.end()) {
				drawnSet.insert(int64f2(ep.x, ep.y));
	            draw_subdiv_facet( src, dst, cvSubdiv2DRotateEdge( e, 3 ), doFill, doDrawLine);
			}
        }

        CV_NEXT_SEQ_ELEM( elem_size, reader );
    }
}

static void drawSubdivDelaunay(IplImage *src, IplImage *dst, CvSubdiv2D* subdiv, bool doFill = true, bool doDrawLine = true)
{
    CvSeqReader  reader;
    int i, total = subdiv->edges->total;
    int elem_size = subdiv->edges->elem_size;
	std::set<CvSubdiv2DEdge> drawnSet;

    cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );

    for( i = 0; i < total; i++ )
    {
        CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);

        if( CV_IS_SET_ELEM( edge ))
        {
            CvSubdiv2DEdge oe = (CvSubdiv2DEdge)edge;

			if (drawnSet.find(oe) == drawnSet.end()) {
				draw_subdiv_facet( src, dst, oe, doFill, doDrawLine);
				CvSubdiv2DEdge se = oe;
				CvSubdiv2DEdge e  = se;
				do {
					drawnSet.insert(e);
					e = cvSubdiv2DGetEdge(e, CV_NEXT_AROUND_LEFT);
				} while (e != se);
			}

            CvSubdiv2DEdge re = cvSubdiv2DRotateEdge(oe, 2);

			if (drawnSet.find(re) == drawnSet.end()) {
				drawnSet.insert(re);
				draw_subdiv_facet( src, dst, re, doFill, doDrawLine);
				CvSubdiv2DEdge se = re;
				CvSubdiv2DEdge e  = se;
				do {
					drawnSet.insert(e);
					e = cvSubdiv2DGetEdge(e, CV_NEXT_AROUND_LEFT);
				} while (e != se);
			}
        }

        CV_NEXT_SEQ_ELEM( elem_size, reader );
    }
}

static void drawSubdiv(IplImage *src, IplImage *dst, CvSubdiv2D* subdiv, bool delaunay, bool doFill = true, bool doDrawLine = true) {
	if (delaunay) {
		drawSubdivDelaunay(src, dst, subdiv, doFill, doDrawLine);
	} else {
		drawSubdivVoronoi(src, dst, subdiv, doFill, doDrawLine);
	}
}


/*
static void drawVoronoi(IplImage *src, IplImage *dst)
{
    char win[] = "source";
    int i;
    CvRect rect = { 0, 0, src->width, src->height };
    CvMemStorage* storage;
    CvSubdiv2D* subdiv;
	CvScalar active_facet_color, delaunay_color, voronoi_color, bkgnd_color;

    active_facet_color = CV_RGB( 255, 0, 0 );
    delaunay_color  = CV_RGB( 0,0,0);
    voronoi_color = CV_RGB(0, 180, 0);
    bkgnd_color = CV_RGB(0,255,0);

    cvSet( dst, bkgnd_color, 0 );

    cvNamedWindow( win, 1 );

    storage = cvCreateMemStorage(0);
    subdiv = init_delaunay( storage, rect );

	CvPoint2D32f *fps = (CvPoint2D32f*)malloc(g_nDiv * sizeof(CvPoint2D32f)); 

    for( i = 0; i < g_nDiv; i++ ) {
		fps[i].x = (float)(cvRandReal(&g_rngState) * src->width);
		fps[i].y = (float)(cvRandReal(&g_rngState) * src->height);
        cvSubdivDelaunay2DInsert( subdiv, fps[i] );
    }
    
	cvCalcSubdivVoronoi2D( subdiv );
    cvSet( dst, bkgnd_color, 0 );
    paint_voronoi(src, dst, subdiv);
	
	free(fps);
	
	cvShowImage( win, dst );

    cvReleaseMemStorage( &storage );
}
*/


void releaseSubdiv() {
    cvReleaseMemStorage(&g_subdivStorage);
	g_subdiv = NULL;
}

void updateSubdiv(int width, int height) {
	releaseSubdiv();
	g_subdivStorage = cvCreateMemStorage(0);

	CvRect rect = {0, 0, width, height};
    g_subdiv = init_delaunay(g_subdivStorage, rect );

	double sin60 = sqrt(3.0) * 0.5;

	if (g_divEqual) {
		double mgn = sqrt((double)(width * height)/(g_nDiv * sin60));
		int xdiv = Saturate(cvRound(width/mgn), 0, width);
		int ydiv = Saturate(cvRound(height/(mgn*sin60)), 0, height);

		float ixdiv = 1.0f/(float)xdiv;
		float iydiv = 1.0f/(float)ydiv;

		float fw = (float)width;
		float fh = (float)height;
		float wd = fw*ixdiv;
		float hd = fh*iydiv;

		for (int j = 0; j <= ydiv; j++) {
			for (int i = 0; i <= xdiv; i++) {
				float d = (j % 2) * wd * 0.5f; 
				CvPoint2D32f pt = {wd*i + d, hd*j};
				pt.x += (float)(wd * (cvRandReal(&g_rngState) - 0.5) * (1.0-g_equality));
				pt.y += (float)(hd * (cvRandReal(&g_rngState) - 0.5) * (1.0-g_equality));
				pt.x = Saturate(pt.x, 0.0f, (float)(width-1));
				pt.y = Saturate(pt.y, 0.0f, (float)(height-1));
				cvSubdivDelaunay2DInsert(g_subdiv, pt);
			}
		}
	} else {
		for(int i = 0; i < g_nDiv; i++ ) {
			CvPoint2D32f pt;
			pt.x = (float)(cvRandReal(&g_rngState) * width);
			pt.y = (float)(cvRandReal(&g_rngState) * height);
			pt.x = Saturate(pt.x, 0.0f, (float)(width-1));
			pt.y = Saturate(pt.y, 0.0f, (float)(height-1));
			cvSubdivDelaunay2DInsert(g_subdiv, pt);
	    }
	}
    
	cvCalcSubdivVoronoi2D(g_subdiv);	
}


static inline int
icvIsRightOf( CvPoint2D32f& pt, CvSubdiv2DEdge edge )
{
    CvSubdiv2DPoint *org = cvSubdiv2DEdgeOrg(edge), *dst = cvSubdiv2DEdgeDst(edge);
    Cv32suf cw_area;
    cw_area.f = (float)cvTriangleArea( pt, dst->pt, org->pt );

    return (cw_area.i > 0)*2 - (cw_area.i*2 != 0);
}

static inline int
icvIsRightOf2( const CvPoint2D32f& pt, const CvPoint2D32f& org, const CvPoint2D32f& diff )
{
    Cv32suf cw_area;
    cw_area.f = (org.x - pt.x)*diff.y - (org.y - pt.y)*diff.x;
    return (cw_area.i > 0)*2 - (cw_area.i*2 != 0);
}

static
CvSubdiv2DPoint*
findNearestPtEdge( CvSubdiv2D* subdiv, CvPoint2D32f pt, CvSubdiv2DEdge* ed)
{
    CvSubdiv2DPoint* point = 0;
    CvPoint2D32f start;
    CvPoint2D32f diff;
    CvSubdiv2DPointLocation loc;
    CvSubdiv2DEdge edge; 
    int i;
    
    CV_FUNCNAME("findNearestPtEdge");

    __BEGIN__;

    if( !subdiv )
        CV_ERROR( CV_StsNullPtr, "" );

    if( !CV_IS_SUBDIV2D( subdiv ))
        CV_ERROR( CV_StsNullPtr, "" );

    if( !subdiv->is_geometry_valid )
        cvCalcSubdivVoronoi2D( subdiv );

    loc = cvSubdiv2DLocate( subdiv, pt, &edge, &point );

    switch( loc )
    {
    case CV_PTLOC_ON_EDGE:
    case CV_PTLOC_INSIDE:
        break;
    default:
        EXIT;
    }

    point = 0;

    start = cvSubdiv2DEdgeOrg( edge )->pt;
    diff.x = pt.x - start.x;
    diff.y = pt.y - start.y;

    edge = cvSubdiv2DRotateEdge( edge, 1 );

    for( i = 0; i < subdiv->total; i++ )
    {
        CvPoint2D32f t;
        
        for(;;)
        {
            assert( cvSubdiv2DEdgeDst( edge ));
            
            t = cvSubdiv2DEdgeDst( edge )->pt;
            if( icvIsRightOf2( t, start, diff ) >= 0 )
                break;

            edge = cvSubdiv2DGetEdge( edge, CV_NEXT_AROUND_LEFT );
        }

        for(;;)
        {
            assert( cvSubdiv2DEdgeOrg( edge ));

            t = cvSubdiv2DEdgeOrg( edge )->pt;
            if( icvIsRightOf2( t, start, diff ) < 0 )
                break;

            edge = cvSubdiv2DGetEdge( edge, CV_PREV_AROUND_LEFT );
        }

        {
            CvPoint2D32f tempDiff = cvSubdiv2DEdgeDst( edge )->pt;
            t = cvSubdiv2DEdgeOrg( edge )->pt;
            tempDiff.x -= t.x;
            tempDiff.y -= t.y;

            if( icvIsRightOf2( pt, t, tempDiff ) >= 0 )
            {
				*ed = edge;
                point = cvSubdiv2DEdgeOrg( cvSubdiv2DRotateEdge( edge, 3 ));
                break;
            }
        }

        edge = cvSubdiv2DSymEdge( edge );
    }

    __END__;

    return point;
}

static void fillSubdivFacet(IplImage *dst, CvSubdiv2DEdge edge, CvScalar col) {
    CvSubdiv2DEdge t = edge;
    int i, count = 0;
	CvPoint ptbuf[255];
	CvPoint *buf = ptbuf;

    // count number of edges in facet
    do
    {
        count++;
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    } while (t != edge );

	if (count > 255) {
		return;
	}

    // gather points
    t = edge;
    for( i = 0; i < count; i++ )
    {
        CvSubdiv2DPoint* pt = cvSubdiv2DEdgeOrg( t );
        if( !pt ) break;
		buf[i] = cvPoint( cvRound(pt->pt.x), cvRound(pt->pt.y));
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    }

    if( i == count ) {
		cvFillConvexPoly( dst, buf, count, col, g_lineType, 0 );
    }
}

void makeMaskSubdiv(IplImage *src, IplImage *dst, CvSubdiv2D* subdiv, bool delaunay) {
	int w = src->width;
	int h = src->height;

	if (!delaunay) {
		typedef std::map<CvSubdiv2DPoint *, CvSubdiv2DEdge> Pt2EdgeMap;
		Pt2EdgeMap pt2edge;
		std::set<int64> drawnSet;

		for(int iy = 0; iy < h; iy++) {
			for(int ix = 0; ix < w; ix++) {
				if (CV_IMAGE_ELEM(src, BYTE, iy, ix) <= 127) {
					continue;
				}
				if (CV_IMAGE_ELEM(dst, BYTE, iy, ix) == 255) {
					int ixm1 = std::max<int>(ix-1, 0);
					int iym1 = std::max<int>(iy-1, 0);
					int ixp1 = std::min<int>(ix+1, w-1);
					int iyp1 = std::min<int>(iy+1, h-1);

					if (CV_IMAGE_ELEM(dst, BYTE, iym1, ixm1) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iym1, ix  ) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iym1, ixp1) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iy,   ixm1) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iy,   ixp1) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iyp1, ixm1) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iyp1, ix  ) == 255 &&
						CV_IMAGE_ELEM(dst, BYTE, iyp1, ixp1) == 255) {
						continue;
					}
				}

				CvPoint2D32f pt = {(float)ix, (float)iy};
				CvSubdiv2DEdge edge = NULL;
				CvSubdiv2DPoint *sp = cvFindNearestPoint2D(subdiv, pt);

				if (drawnSet.find(int64f2(sp->pt.x, sp->pt.y)) != drawnSet.end()) {
					continue;
				}
				drawnSet.insert(int64f2(sp->pt.x, sp->pt.y));

				Pt2EdgeMap::iterator itr = pt2edge.find(sp);
				if (itr != pt2edge.end()) {
					edge = itr->second;
				} else {
					CvSubdiv2DPoint *sp = findNearestPtEdge(subdiv, pt, &edge);
					pt2edge[sp] = edge;
				}

				if (edge == NULL) {
					continue;
				}

				fillSubdivFacet(dst, edge, CV_RGB(255,255,255));
			}
		}
	} else {
		std::set<CvSubdiv2DEdge> drawnSet;

		for(int iy = 0; iy < h; iy++) {
			for(int ix = 0; ix < w; ix++) {
				if (CV_IMAGE_ELEM(src, BYTE, iy, ix) <= 127) {
					continue;
				}
				if (CV_IMAGE_ELEM(dst, BYTE, iy, ix) == 255) {
					continue;
				}

				CvPoint2D32f pt = {(float)ix, (float)iy};
				CvSubdiv2DPoint* vpt = 0;
				CvSubdiv2DEdge edge; 
				CvSubdiv2DPointLocation loc;
				loc = cvSubdiv2DLocate(subdiv, pt, &edge, &vpt);

				if (loc != CV_PTLOC_ON_EDGE && loc != CV_PTLOC_INSIDE) {
					CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+0) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+0);
					CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+1) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+1);
					CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+2) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+2);
					continue;
				}

				if (drawnSet.find(edge) == drawnSet.end()) {
					fillSubdivFacet(dst, edge, CV_RGB(255,255,255));
					CvSubdiv2DEdge se = edge;
					CvSubdiv2DEdge e  = se;
					do {
						drawnSet.insert(e);
						e = cvSubdiv2DGetEdge(e, CV_NEXT_AROUND_LEFT);
					} while (e != se);
				}
			}
		}
	}

	return;
}


void test1(IplImage *src, IplImage *dst) {
	//IplImage *dst = cvCreateImage(cvSize(src->width, src->height), IPL_DEPTH_8U, 3);
    CvScalar bkgnd_color = CV_RGB(0,255,0);
	cvSet( dst, bkgnd_color, 0 );

	typedef std::map<CvSubdiv2DPoint *, CvSubdiv2DEdge> Pt2EdgeMap;
	Pt2EdgeMap pt2edge;

	int w = src->width;
	int h = src->height;

	for(int iy = 0; iy < h; iy++) {
		for(int ix = 0; ix < w; ix++) {
			CvPoint2D32f pt = {(float)ix, (float)iy};
#if 1
			CvSubdiv2DEdge edge = NULL;
			CvSubdiv2DPoint *sp = cvFindNearestPoint2D(g_subdiv, pt);
			Pt2EdgeMap::iterator itr = pt2edge.find(sp);
			if (itr != pt2edge.end()) {
				edge = itr->second;
			} else {
				CvSubdiv2DPoint *sp = findNearestPtEdge(g_subdiv, pt, &edge);
				pt2edge[sp] = edge;
			}

			if (edge == NULL) {
				int x = Saturate(cvRound(sp->pt.x), 0, src->width - 1);
				int y = Saturate(cvRound(sp->pt.y), 0, src->height - 1);

				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+0) = CV_IMAGE_ELEM(src, BYTE, y, x*3+0);
				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+1) = CV_IMAGE_ELEM(src, BYTE, y, x*3+1);
				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+2) = CV_IMAGE_ELEM(src, BYTE, y, x*3+2);
				continue;
			}

			CvPoint2D32f bpt1f = {0.0,0.0};

			float ixf = (float)ix;
			float iyf = (float)iy;

			float b = 0;
			float g = 0;
			float r = 0;
			float dist = 0;

			CvSubdiv2DEdge t = edge;
			int n = 0;
			do {
				CvSubdiv2DPoint *org = cvSubdiv2DEdgeOrg(t);
				if (org == NULL) break;

				CvPoint2D32f pt1f = org->pt;

				float d = (pt1f.x-ixf)*(pt1f.x-ixf) + (pt1f.y-iyf)*(pt1f.y-iyf);
				d = 1.0f/d;
				/*
				if (d < 0.1) {
					d = 10.0*10.0;
				} else {
					d = 1.0/(d*d);
				}
				*/

				int x = Saturate((int)pt1f.x, 0, src->width - 1);
				int y = Saturate((int)pt1f.y, 0, src->height - 1);

				b += d * CV_IMAGE_ELEM(src, BYTE, y, x*3+0);
				g += d * CV_IMAGE_ELEM(src, BYTE, y, x*3+1);
				r += d * CV_IMAGE_ELEM(src, BYTE, y, x*3+2);
				dist += d;

				/*
				if (ix == 320 && iy == 200) {
					if (n >= 1) {
						CvPoint pt1 = {cvRound(bpt1f.x), cvRound(bpt1f.y)};
						CvPoint pt2 = {cvRound(pt1f.x), cvRound(pt1f.y)};
						cvLine(dst, pt1, pt2, CV_RGB(255,0,0));
						draw_subdiv_point(dst, pt, CV_RGB(0,0,255));
					} else {
						draw_subdiv_point(dst, pt1f, CV_RGB(0,255,255));
					}
				}
				bpt1f = org->pt;
				*/

				t = cvSubdiv2DGetEdge(t, CV_NEXT_AROUND_LEFT);

				n++;
			} while (t != edge);

			b /= dist;
			g /= dist;
			r /= dist;

			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+0) = (BYTE)b;
			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+1) = (BYTE)g;
			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+2) = (BYTE)r;
#else
			CvSubdiv2DPoint* vpt = 0;
			CvSubdiv2DEdge edge; 
			CvSubdiv2DPointLocation loc;
			loc = cvSubdiv2DLocate(g_subdiv, pt, &edge, &vpt);

			if (loc != CV_PTLOC_ON_EDGE && loc != CV_PTLOC_INSIDE) {
				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+0) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+0);
				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+1) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+1);
				CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+2) = CV_IMAGE_ELEM(src, BYTE, iy, ix*3+2);
				continue;
			}

			float cb = 0;
			float cg = 0;
			float cr = 0;

			CvSubdiv2DEdge te = edge;
			CvPoint2D32f pt1 = cvSubdiv2DEdgeOrg(te)->pt;
			CvPoint2D32f pt2 = cvSubdiv2DEdgeDst(te)->pt;
			te = cvSubdiv2DGetEdge(te, CV_NEXT_AROUND_LEFT);
			CvPoint2D32f pt3 = cvSubdiv2DEdgeDst(te)->pt;

			CvPoint2D32f p = {pt.x  - pt1.x, pt.y  - pt1.y};
			CvPoint2D32f a = {pt2.x - pt1.x, pt2.y - pt1.y};
			CvPoint2D32f b = {pt3.x - pt1.x, pt3.y - pt1.y};

			float s = (((a.x-b.x)*p.y) - ((a.y-b.y)*p.x)) / (a.x*b.y - a.y*b.x);
			float t = (b.x*p.y - b.y*p.x) / (((b.x-a.x)*p.y) - ((b.y-a.y)*p.x));

			int x,y;
			x = Saturate((int)pt1.x, 0, src->width - 1);
			y = Saturate((int)pt1.y, 0, src->height - 1);
			BYTE b1 = CV_IMAGE_ELEM(src, BYTE, y, x*3+0);
			BYTE g1 = CV_IMAGE_ELEM(src, BYTE, y, x*3+1);
			BYTE r1 = CV_IMAGE_ELEM(src, BYTE, y, x*3+2);

			x = Saturate((int)pt2.x, 0, src->width - 1);
			y = Saturate((int)pt2.y, 0, src->height - 1);
			BYTE b2 = CV_IMAGE_ELEM(src, BYTE, y, x*3+0);
			BYTE g2 = CV_IMAGE_ELEM(src, BYTE, y, x*3+1);
			BYTE r2 = CV_IMAGE_ELEM(src, BYTE, y, x*3+2);

			x = Saturate((int)pt3.x, 0, src->width - 1);
			y = Saturate((int)pt3.y, 0, src->height - 1);
			BYTE b3 = CV_IMAGE_ELEM(src, BYTE, y, x*3+0);
			BYTE g3 = CV_IMAGE_ELEM(src, BYTE, y, x*3+1);
			BYTE r3 = CV_IMAGE_ELEM(src, BYTE, y, x*3+2);

			if (ix == 320 && iy == 200) {
				draw_subdiv_point(dst, pt1, CV_RGB(r1,g1,b1	));
				draw_subdiv_point(dst, pt2, CV_RGB(r2,g2,b2));
				draw_subdiv_point(dst, pt3, CV_RGB(r3,g3,b3));
			}


			cb = (1-s)*b1 + s * (t*b2 + (1-t)*b3);
			cg = (1-s)*g1 + s * (t*g2 + (1-t)*g3);
			cr = (1-s)*r1 + s * (t*r2 + (1-t)*r3);

			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+0) = (BYTE)cb;
			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+1) = (BYTE)cg;
			CV_IMAGE_ELEM(dst, BYTE, iy, ix*3+2) = (BYTE)cr;
#endif
		}
	}
	//cvNamedWindow("test1", CV_WINDOW_AUTOSIZE);
	//cvShowImage("test1", dst);



	//cvReleaseImage(&dst);

	return;
}



BOOL execVoronoi(FILTER *fp,FILTER_PROC_INFO *fpip) {
	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	bool updSubdiv = false;
	int	 b_nDiv = g_nDiv;

	g_nDiv = max(getTrackVal(fp, ""), 1);
	if (g_nDiv != b_nDiv) {
		updSubdiv = true;
	}

	switch (getTrackVal(fp, "LineType")) {
		case 1: g_lineType = 0; break;		//8A
		case 2:	g_lineType = 4; break;		//4A
		case 3:	g_lineType = CV_AA; break;	//A`GCAX
		default: g_lineType = CV_AA; break;
	}

	g_lineL = getTrackVal(fp, "x");
	if (g_lineL >= 0) {
		g_lineColor = CV_RGB(g_lineL, g_lineL, g_lineL);
	} else {
		switch (g_lineL) {
			case -1:
				g_lineColor = CV_RGB(0,255,0);
				break;
			case -2:
				g_lineColor = CV_RGB(0,0,255);
				break;
		}
	}


	double b_equality = g_equality;
	g_equality = (double)getTrackVal(fp, "ϓx")/100.0;
	if (g_equality != b_equality) {
		updSubdiv = true;
	}

	int maskf = getTrackVal(fp, "}XNF");
	int bgf = getTrackVal(fp, "wiF");

	int freq = getTrackVal(fp, "XVԊu");
	int rseed = getTrackVal(fp, "");

	bool doDrawLine = (getCheckVal(fp, "E") == 1);

	bool b_divEqual = g_divEqual;
	g_divEqual = (getCheckVal(fp, "ϓ") == 1);
	if (g_divEqual != b_divEqual) {
		updSubdiv = true;
	}

	bool delaunay = (getCheckVal(fp, "Op`") == 1);


	// KvimagemۂȂ
	if (img0 == NULL) {
		releaseImages();
		createImages(cvSize(w,h));
		updSubdiv = true;
	} else {
		CvSize size = cvGetSize(img0);
		if (size.width != w || size.height != h) {
			releaseImages();
			createImages(cvSize(w,h));
			updSubdiv = true;
		}
	}

	// RNGStateXV
	static int64 seed = -1;
	int64 b_seed = seed;

	if (freq == 0) {
		seed = ((int64)(rseed+g_nDiv)) << 32;
	} else {
		seed = (((int64)(rseed+g_nDiv)) << 32) + fpip->frame / freq;
	}
	g_rngState = cvRNG(seed);

	if (seed != b_seed) {
		updSubdiv = true;
	}

	// {mC
	if (updSubdiv) {
		updateSubdiv(w, h);
	}


	//AviUtl  OpenCV
	CopyBufToCvBGRImg(buffer0, img0, 0,0,w,h, 0,0, mw,mh);
	cvCopy(img0, img1);

	//}XN
	if (maskf >= 0) {
		fp->exfunc->set_ycp_filtering_cache_size(fp, mw, mh, 5, NULL);
		PIXEL_YC *buf_mask = (PIXEL_YC *)fp->exfunc->get_ycp_filtering_cache_ex(fp, fpip->editp, fpip->frame + maskf, NULL, NULL);
		CopyBufToCvGrayImg(buf_mask, img_mask, 0,0,w,h, 0,0, mw,mh);
		cvSet(img_smask, CV_RGB(0,0,0));
		//cvNamedWindow("test1", CV_WINDOW_AUTOSIZE);
		//cvShowImage("test1", img_mask);
	}
	//wi
	if (bgf > 0) {
		fp->exfunc->set_ycp_filtering_cache_size(fp, mw, mh, 5, NULL);
		PIXEL_YC *buf_bg = (PIXEL_YC *)fp->exfunc->get_ycp_filtering_cache_ex(fp, fpip->editp, fpip->frame + bgf, NULL, NULL);
		CopyBufToCvBGRImg(buf_bg, img_bg, 0,0,w,h, 0,0, mw,mh);
	} else if (bgf == -1) {
		cvSet(img_bg, CV_RGB(0,255,0));
	}

	//{mCʕ`
	if (true || getCheckVal(fp, "test") == 0) {
		drawSubdiv(img0, img1, g_subdiv, delaunay, true, false);
		if (doDrawLine) {
			drawSubdiv(img0, img1, g_subdiv, delaunay, false, true);
		}
	} else {
		test1(img0, img1);
		if (doDrawLine) {
			drawSubdiv(img0, img1, g_subdiv, delaunay, false, true);
		}
	}

	IplImage *bg = img1;

	//}XNw肪ꍇ̏
	if (maskf >= 0) {
		makeMaskSubdiv(img_mask, img_smask, g_subdiv, delaunay);
		//cvNamedWindow("test2", CV_WINDOW_AUTOSIZE);
		//cvShowImage("test2", img_smask);

		if (bgf != 0) {
			bg = img_bg;
		} else {
			bg = img0;
		}
		cvErode(img_smask, img_smask);
		cvCopy(img1, bg, img_smask);
	}

	//OpenCV  AviUtl
	CopyCvBGRImgToBuf(bg, buffer0, 0,0,w,h, 0,0, mw,mh);

	return TRUE;
}
